unit main;

interface
//{$DEFINE XBOARD}
{$DEFINE FULL_STATS} 

uses
  Classes, Forms, Dialogs, Messages, SysUtils, StrUtils, DateUtils, Math,
  {$IFDEF XBOARD} WBUnit, {$ENDIF}
  Controls, StdCtrls, Grids, Types, Graphics, PromUnit, Menus, ExtCtrls,
  ComCtrls, ImgList, ToolWin,BookUnit;

const
  LINE_END   = #$D#$A;
  INFINITY   = 30000;
  MATE_VALUE = 10000;
  CAPT_SORT  = 10000000;
  MAX_DEPTH  = 36;       // safety limit for search depth
  MOVE_LIMIT = 800;      // maximal stack size
  BOOK_SIZE  = 100000;
  COMM_SIZE  = 5000;     // max no. of opening book comments
  HASH_SIZE  = 262144 * 8;

// piece types
PAWN = 1; KNIGHT = 2; BISH = 3; ROOK = 4; QUEEN = 5; KING = 6;

OUTSIDE = 120; // square outside of the board
WHITE   = 1;      BLACK   = 0;

NO_PROM = 9;
END_MAT = -12;

// square numbers according to 0x88 scheme
A8=112; B8=113; C8=114; D8=115; E8=116; F8=117; G8=118; H8=119;
A7=96;  B7=97;  C7=98;  D7=99;  E7=100; F7=101; G7=102; H7=103;
A6=80;  B6=81;  C6=82;  D6=83;  E6=84;  F6=85;  G6=86;  H6=87;
A5=64;  B5=65;  C5=66;  D5=67;  E5=68;  F5=69;  G5=70;  H5=71;
A4=48;  B4=49;  C4=50;  D4=51;  E4=52;  F4=53;  G4=54;  H4=55;
A3=32;  B3=33;  C3=34;  D3=35;  E3=36;  F3=37;  G3=38;  H3=39;
A2=16;  B2=17;  C2=18;  D2=19;  E2=20;  F2=21;  G2=22;  H2=23;
A1=0;   B1=1;   C1=2;   D1=3;   E1=4;   F1=5;   G1=6;   H1=7;

// Row and column identifiers
ROW_1 = A1 div 16; ROW_2 = A2 div 16; ROW_3 = A3 div 16; ROW_4 = A4 div 16;
ROW_5 = A5 div 16; ROW_6 = A6 div 16; ROW_7 = A7 div 16; ROW_8 = A8 div 16;
COL_A = A1 mod 16; COL_B = B1 mod 16; COL_C = C1 mod 16; COL_D = D1 mod 16;
COL_E = E1 mod 16; COL_F = F1 mod 16; COL_G = G1 mod 16; COL_H = H1 mod 16;

// slots in the empty part of the 0x88 board
WKMoved = 8;   WKRMoved = 10;  WQRMoved = 11;
BKMoved = 9;   BKRMoved = 12;  BQRMoved = 13;
WPCount = 14;  BPCount  = 15;  // no. of pawns for each side
WNCount = 24;  BNCount  = 25;  // no. of knights           .
WBCount = 26;  BBCount  = 27;  // no. of bishops           .
WRCount = 28;  BRCount  = 29;  // no. of rooks             .
WQCount = 30;  BQCount  = 31;  // no. of queens            .
WMat    = 40;  BMat     = 41;  // artificial value for PIECE material

// vectors
NORTH = 16;           NN    = NORTH + NORTH;
SOUTH = -16;          SS    = SOUTH + SOUTH;
EAST  = 1;            WEST  = -1;
NE    = 17;           SW    = -17;
NW    = 15;           SE    = -15;
NNE   = NN + EAST;    NNW   = NN + WEST;
SSE   = SS + EAST;    SSW   = SS + WEST;

// move types
NORM     = 1;  PAWNJUMP = 2;   CAPT  = 3;  PROM = 4;
CAPTPROM = 5;  EPCAPT   = 6;  CASTLE = 7;  NOFLAG   = 8;

EXACT = 1; BELOW_ALFA = 2; ABOVE_BETA = 3; // hash flags

B_NORM = 0;  B_BAD  = 1;  B_FUN  = 2;      // book flags
B_EXCL = 3;  B_DISL = 4;  B_END = 5;

type
  TLang = ( POLISH, ENGLISH );
  TInd  = -16..16;
  TDat  = array[ 0..127 ] of Integer;
  TPieceVal   = array[ 0..6 ] of Integer;
  TDispArray  = array[ 0..7, 0..7 ] of Integer;
  TCoef       = array[ 0..127, 0..127 ] of Integer;
  TEndgame    = ( UNDEFINED, PAWN_END, BISH_END, BOC_END,
                KNIGHT_END, BN_END, KRPKR, KBPKX, ROOK_END );

  THashRec = record
     Key    : Integer;
     Depth  : ShortInt;
     SqFrom : ShortInt;
     SqTo   : ShortInt;
     Val    : SmallInt;
     Flag   : Byte;
  end; // THashRec

  TSmallHashRec = record
     Key : Integer;
     Val : SmallInt;
  end; // TSmallHashRec

  TFullBook = record
     Fen    :  string[ 78 ];
     Move   :  string[ 4 ];
     Freq   :  Integer;       // frequency of move input
     Learn  :  Integer;       // user-defined move value
  end;  // TFullBook

  TPosComment = record
     Fen      : string[ 78 ];
     Comment  : string[ 255 ];
  end;

  TPiece = record             // In TPiece "Kind" is piece
    Kind  : TInd;             // kind,  "Sqr" its location
    Sqr   : ShortInt;         // on the board ( or OUTSIDE
    Color : TInd;             // if a piece has been  cap-
  end; // TPiece              // tured ). "Color"  is  set
                              // at program initialization
  // and  never  changed  afterwards.  Though  it  can  be
  // inferred from the location on the piece list ( if < 0
  // then  a  piece  is black ), the  additional  variable
  // simplifies the move generation.

  TList      = array[ -16..16 ] of TPiece;
  THash      = array[ 0..HASH_SIZE ] of THashRec;
  TSmallHash = array[ 0..HASH_SIZE ] of TSmallHashRec;
  THashData  = array[ 0..127, 0..1, 1..6 ] of Integer;
  THashFile  = file of THashData;

  TBoard = record
    Index : array[ 0..127 ] of TInd; // piece location on the list or 0
    List  : TList;                   // piece list
    Side  : ShortInt;                // side to move
    Ep    : ShortInt;                // en passant square
    MovesTo50 : ShortInt;            // counter for 50 move rule
  end; // TBoard

  TMove = record
    SqFr, SqTo : Byte; // move coordinates                 /
    Prom   : TInd;     // promoted piece                   /
    Flag   : TInd;     // move kind, see constants         /
    Order  : Integer;  // value used for sorting           /
  end; // TMove

  TMoveSet  = array [ 1..250 ] of TMove;
  TVect     = array [ 1..8 ] of Integer;
  TFileStat = array [ -1..9, 0..1 ] of TInd;
  TTiming   = ( SEC_PER_MOVE, FIXED_DEPTH, MIN_PER_GAME, PERIODS );
  TNotation = ( N_SAN, N_PGN );

  {$I header.pas} // class definitions are stored there

  TForm1 = class( TForm )
    BoardGrid: TStringGrid;
    ColorDialog: TColorDialog;
    Timer1: TTimer;
    MainMenu1: TMainMenu;

    Game1   : TMenuItem;     iming1: TMenuItem;
    Depth1  : TMenuItem;     Permove1: TMenuItem;
    Pergame1: TMenuItem;

    N1sec1 : TMenuItem;      N5sec1 : TMenuItem;
    N10sec1: TMenuItem;      N30sec1: TMenuItem;
    N60sec1: TMenuItem;      N120s1 : TMenuItem;

    N4plies1 : TMenuItem;    N5plies1 : TMenuItem;
    N6plies1 : TMenuItem;    N7plies1 : TMenuItem;
    N8plies1 : TMenuItem;    N9plies1 : TMenuItem;
    N10plies1: TMenuItem;    N12plies1: TMenuItem;

    N5min1   : TMenuItem;    N10min1: TMenuItem;
    N15min1  : TMenuItem;    N30min1: TMenuItem;
    N1h1     : TMenuItem;    N15h1  : TMenuItem;
    N2h1     : TMenuItem;

    AnMemo   : TMemo;           GameMemo    : TMemo;
    New1     : TMenuItem;       Quit1       : TMenuItem;
    Viev1    : TMenuItem;       Statistics1 : TMenuItem;
    hink1    : TMenuItem;       Deactivate1 : TMenuItem;
    Invert1  : TMenuItem;       Colors1     : TMenuItem;
    Lightsquared1: TMenuItem;   Darksquares1: TMenuItem;
    Whitepieces1: TMenuItem;    Blackpieces1: TMenuItem;
    ools1: TMenuItem;           BookEditor1 : TMenuItem;

    WPName    : TLabel;         BPName      : TLabel;
    WMoveTime : TLabel;         BmoveTime   : TLabel;
    WTime     : TLabel;         BTime       : TLabel;

    Bevel1     : TBevel;        ToolBar1: TToolBar;
    ImageList1 : TImageList;    Image1  : TImage;

    BackButt   : TToolButton;   BackAllButt: TToolButton;
    FwdButt    : TToolButton;   FwdAllButt : TToolButton;
    BookButt   : TToolButton;   QuitButt   : TToolButton;
    AnalyzeButt: TToolButton;   InfoButt   : TToolButton;
    EvalButt   : TToolButton;   ThinkButt  : TToolButton;
    NewGameButt: TToolButton;   StopButt   : TToolButton;
    StatButt   : TToolButton;   InvertButt : TToolButton;

    Settings1    : TMenuItem;   Boardbackground1 : TMenuItem;
    Savecurrent1 : TMenuItem;   Restoreoriginal1 : TMenuItem;
    Lightfont1   : TMenuItem;   Darkfont1 : TMenuItem;
    Windowfont1  : TMenuItem;   WindowBackground1 : TMenuItem;

    Playername1: TMenuItem;     EvalEditor1: TMenuItem;
    WhiteClockBevel: TBevel;    BlackClockBevel: TBevel;
    InfoMemo: TMemo;

    Bookstyle1   : TMenuItem;   Broad1   : TMenuItem;
    Normal1      : TMenuItem;   Solid1   : TMenuItem;
    Hypermodern1 : TMenuItem;   Gambit1  : TMenuItem;
    Fighter1     : TMenuItem;   Unusual1 : TMenuItem;
    Bookoff1     : TMenuItem;

    Lauouts1     : TMenuItem;   Subdued1 : TMenuItem;
    YellowWater1 : TMenuItem;   Glacier1 : TMenuItem;
    Wood1        : TMenuItem;   undra1   : TMenuItem;

    Language1    : TMenuItem;
    Polish1      : TMenuItem;   English1 : TMenuItem;
    Playingstyle1: TMenuItem;
    Objective1   : TMenuItem;   Smart1: TMenuItem;
    Pessimist1   : TMenuItem;
    SearchEditor1: TMenuItem;   Dumptree1    : TMenuItem;
    est1         : TMenuItem;   Genspeed1    : TMenuItem;

    procedure BoardGridDrawCell( Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState );
    procedure InvertPieceStruct;
    procedure BoardGridMouseDown( Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer );
    procedure BoardGridMouseUp( Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer );
    procedure N60sec1Click( Sender: TObject );
    procedure N30sec1Click( Sender: TObject );
    procedure N10sec1Click( Sender: TObject );
    procedure N5sec1Click( Sender: TObject );
    procedure N1sec1Click( Sender: TObject );
    procedure N4plies1Click( Sender: TObject );
    procedure N5plies1Click( Sender: TObject );
    procedure N6plies1Click( Sender: TObject );
    procedure N7plies1Click( Sender: TObject );
    procedure N8plies1Click( Sender: TObject );
    procedure N9plies1Click( Sender: TObject );
    procedure N10plies1Click( Sender: TObject );
    procedure N5min1Click( Sender: TObject );
    procedure N10min1Click( Sender: TObject );
    procedure N15min1Click( Sender: TObject );
    procedure N30min1Click( Sender: TObject );
    procedure N1h1Click( Sender: TObject );
    procedure N15h1Click( Sender: TObject );
    procedure N2h1Click( Sender: TObject );
    procedure New1Click( Sender: TObject );
    procedure Quit1Click( Sender: TObject );
    procedure Statistics1Click( Sender: TObject );
    procedure hink1Click( Sender: TObject );
    procedure Invert1Click( Sender: TObject );
    procedure Deactivate1Click( Sender: TObject );
    procedure Lightsquared1Click( Sender: TObject );
    procedure Darksquares1Click( Sender: TObject );
    procedure Whitepieces1Click( Sender: TObject );
    procedure Blackpieces1Click( Sender: TObject );
    procedure Timer1Timer( Sender: TObject );
    procedure BackButtClick( Sender: TObject );
    procedure BackAllButtClick( Sender: TObject );
    procedure FwdButtClick( Sender: TObject );
    procedure FwdAllButtClick( Sender: TObject );
    procedure BookEditor1Click( Sender: TObject );
    procedure AnalyzeButtClick( Sender: TObject );
    procedure InfoButtClick( Sender: TObject );
    procedure EvalButtClick( Sender: TObject );
    procedure Savecurrent1Click(Sender: TObject);
    procedure Restoreoriginal1Click(Sender: TObject);
    procedure Lightfont1Click(Sender: TObject);
    procedure Darkfont1Click(Sender: TObject);
    procedure Windowfont1Click(Sender: TObject);
    procedure Windowbackground1Click(Sender: TObject);
    procedure Playername1Click(Sender: TObject);
    procedure EvalEditor1Click(Sender: TObject);
    procedure Boardbackground1Click(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure Normal1Click(Sender: TObject);
    procedure Broad1Click(Sender: TObject);
    procedure Solid1Click(Sender: TObject);
    procedure Hypermodern1Click(Sender: TObject);
    procedure Fighter1Click(Sender: TObject);
    procedure Gambit1Click(Sender: TObject);
    procedure Unusual1Click(Sender: TObject);
    procedure N12plies1Click(Sender: TObject);
    procedure N120s1Click(Sender: TObject);
    procedure Yellowwater1Click(Sender: TObject);
    procedure Glacier1Click(Sender: TObject);
    procedure Wood1Click(Sender: TObject);
    procedure undra1Click(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure Polish1Click(Sender: TObject);
    procedure English1Click(Sender: TObject);
    procedure Subdued1Click(Sender: TObject);
    procedure Objective1Click( Sender: TObject );
    procedure Smart1Click( Sender: TObject );
    procedure Pessimist1Click( Sender: TObject );

    procedure FormCreate ( Sender: TObject );
    procedure FormDestroy( Sender: TObject );
    procedure SearchEditor1Click(Sender: TObject);
    procedure Dumptree1Click(Sender: TObject);
    procedure Bookoff1Click(Sender: TObject);
    procedure Genspeed1Click(Sender: TObject);

private
    MoveMade  : Boolean;
    {$IFDEF XBOARD} WBThread  : TWBThread; {$ENDIF}

    procedure BookSelection( CurrentBook : string );
    procedure LanguageSettings;
    procedure UpdateWindowOptions;
    procedure UpdateNotation;
    function  GameToSan : string;
    function  MoveToPGN( Move : TMove ) : string;
    function  GameToPGN : string;
    procedure SetSeconds( Ex : string );
    procedure SetDepth( Ex : string );
    procedure SetTime( Ex : string );
    procedure RefreshBoardDisplay;
    procedure BeforeActualMove( Move : TMove );
    procedure ProcessInitString( CurrentString : string);
    procedure ReadIniFile( FileName : string );
    procedure WriteIniFile( FileName : string );
    procedure RestoreSettings;
    procedure SetProgSide;
public
    InputMode : ( imDragging, imClicking );
    IsDumping : Boolean;

    // Customizable values for evaluation function
    ProgMobility  : Integer;      OppoMobility  : Integer;
    ProgAttack    : Integer;      OppoAttack    : Integer;
    ProgShield    : Integer;      OppoShield    : Integer;

    procedure InfoMessage( What : string );
    procedure AddToAnMemo( What : string );
    procedure FenToGraphBoard( ReadString : string );
    procedure PolishLanguage;
    procedure EnglishLanguage;
    procedure GraphBoardToInternal;
    function BoardToFen : string;

    procedure Execute( Command : string );
    procedure AfterActualMove( Move : TMove );
    {$IFDEF XBOARD} procedure WinboardMessage( var Msg : TMessage ); message CW_WBMSG; {$ENDIF}
end; // TMainForm

var
  // Text strings - made variables for translation purposes

  sProgramInfo : string = 'Clericus 1.03' + LINE_END
                       + 'a chess program by Pawel Koziol'
                       + LINE_END + 'nescitus@o2.pl';
  sDrawByRep   : string = 'Draw' + LINE_END + 'by repetition';
  sDrawBy50m   : string = 'Draw' + LINE_END + '50 move rule';
  sDrawByMat   : string = 'Draw' + LINE_END + 'insufficient material';
  sCheckmate   : string = 'Checkmate!';
  sStalemate   : string = 'Stalemate!';
  sBookSaving  : string = 'Saving book...';
  sBookSaved   : string = 'Book saved';
  sMoveBack    : string = 'Taking back';
  sMoveFwd     : string = 'Replaying';
  sFwdAll      : string = 'Go to end';
  sBackAll     : string = 'Go to start';

  IniFile  : TextFile;
  MyStat   : TStat;
  History  : THistory;
  MyEval   : TEval;
  MyBook   : TBook;
  MySearch : TSearch;

  Language    : TLang   = POLISH;
  ClockSide   : Integer = WHITE;
  ProgSide    : Integer = BLACK;
  SearchDepth : Integer = 7;
  Want_resign : Integer = 0;
  ProgPassive : Boolean = False;
  ProgThinking: Boolean = False;
  LastLine    : string;
  LastVal     : Integer;
  WMTime, BMTime, WTotTime, BTotTime : Integer;

// graphical settings
  LtSqrCol      : Integer = clSilver;
  DarkSqrCol    : Integer = clGray;
  WhitePieceCol : Integer = clWhite;
  BlackPieceCol : Integer = clBlack;
  DarkFontCol   : Integer = clBlack;
  LightFontCol  : Integer = clGray;
  WindowFontCol : Integer = clBlack;
  WindowBgCol   : Integer = clWhite;
  BoardBgColor  : Integer = clSilver;
  PictureString : string  = 'graf/marble3.bmp';
  Notation      : TNotation = N_PGN;
  PlayerName    : string = 'Player';
  ProgName      : string = 'Clericus';

// engine settings
  MainBookName     : string = 'books/MainBook.hop';
  GuideBookName    : string = 'books/GuideBook.hop';
  BookName         : string;
  CommentName      : string = 'books/Comments.eng';
  RESIGN_VALUE     : Integer = -900;
  RESIGN_BOARD_VAL : Integer = -500;
  RESIGN_DURATION  : Integer = 3;

  BoardSqrCol, PieceCol  : TDispArray;
  StartCol, StartRow, EndCol, EndRow : Integer;
  ColName  : array[ 0..7 ] of char = ('a','b','c','d','e','f','g','h');
  RowName  : array[ 0..7 ] of char = ('8','7','6','5','4','3','2','1');

// auxiliary
KnightArray : TVect = ( 18, -18, 33, -33, 31, -31, 14, -14 );
StarArray   : TVect = ( NORTH, SOUTH, EAST, WEST, NE, SE, NW, SW );
KingEnPrise : Boolean = False;
ThinkLine   : array[ -36..36 ] of string;
FileStat, EmptyFiles    : TFileStat;

HashData : THashData;   // array of values used for hashing
RandFile : THashFile;   // file with those values
MainHash : THash;       // main hash table

// book variables
BookNo        : Integer = 0;
NofEntries    : Integer;
NofComments   : Integer;
BookArray     : array[ 0..1, 0..BOOK_SIZE ] of TFullBook;
CommentArray  : array[ 0..BOOK_SIZE ] of TPosComment;
BookFile      : file of TFullBook;
CommentFile   : file of TPosComment;
OpeningMove   : TMove;
MissedOpening : Integer = 0;

Board      : TBoard;
Command    : string;
GameStack  : array[ 0..MOVE_LIMIT ] of TBoard;
MoveStack  : array[ 0..MOVE_LIMIT ] of TMove;
RepStack   : array[ 0..MOVE_LIMIT ] of Integer;
StackIndex : Integer = 0;
StackBound : Integer = 0;
MyMove     : TMove;

MoveNo   : Integer;                   // variables used
MoveSet  : TMoveSet;                  // for move generation

Timing   : TTiming = MIN_PER_GAME;    // variables used
STi, ETi : TDateTime;                 // for timing
SecUsed, TimeUsed  : Integer;
AddSec    :  Integer;
MoveTime  : Integer = 10;
BaseTime  : Integer = 300;
BaseInc   : Integer;
BaseMoves, MemMoves : Integer;
MemTime   : Integer;
TimeDiv   : Integer = 30;
NoMem     : Boolean = False;

// variables dealing with conversion form and to string format
WPieceChar  : array[ 0..6 ] of char = ('.','P','N','B','R','Q','K');
BPieceChar  : array[ 0..6 ] of char = ('.','p','n','b','r','q','k');

SqrName : array[ 0..127 ] of string =
         ( 'a1','b1','c1','d1','e1','f1','g1','h1',  '*','x','x','x','x','x','x','x',
           'a2','b2','c2','d2','e2','f2','g2','h2',  'x','x','x','x','x','x','x','x',
           'a3','b3','c3','d3','e3','f3','g3','h3',  'x','x','x','x','x','x','x','x',
           'a4','b4','c4','d4','e4','f4','g4','h4',  'x','x','x','x','x','x','x','x',
           'a5','b5','c5','d5','e5','f5','g5','h5',  'x','x','x','x','x','x','x','x',
           'a6','b6','c6','d6','e6','f6','g6','h6',  'x','x','x','x','x','x','x','x',
           'a7','b7','c7','d7','e7','f7','g7','h7',  'x','x','x','x','x','x','x','x',
           'a8','b8','c8','d8','e8','f8','g8','h8',  '*','x','x','x','x','x','x','x');

SqrSym : TDat = (H1, G1, F1, E1, D1, C1, B1, A1,  0, 0, 0, 0, 0, 0, 0, 0,
                 H2, G2, F2, E2, D2, C2, B2, A2,  0, 0, 0, 0, 0, 0, 0, 0,
                 H3, G3, F3, E3, D3, C3, B3, A3,  0, 0, 0, 0, 0, 0, 0, 0,
                 H4, G4, F4, E4, D4, C4, B4, A4,  0, 0, 0, 0, 0, 0, 0, 0,
                 H5, G5, F5, E5, D5, C5, B5, A5,  0, 0, 0, 0, 0, 0, 0, 0,
                 H6, G6, F6, E6, D6, C6, B6, A6,  0, 0, 0, 0, 0, 0, 0, 0,
                 H7, G7, F7, E7, D7, C7, B7, A7,  0, 0, 0, 0, 0, 0, 0, 0,
                 H8, G8, F8, E8, D8, C8, B8, A8,  0, 0, 0, 0, 0, 0, 0, 0);

ArrToSq : array [0..7, 0..7] of Integer;

{$I params.pas}   // evaluation function parameters

Form1: TForm1;

implementation

uses names, EvalEditor, DebugWindow, SearchEd, DumpUnit;
  {$R *.dfm}

procedure TForm1.InfoMessage( What : string );
begin
   InfoMemo.Font.Size := 16;
   InfoMemo.Text := What;
   InfoMemo.Repaint;
end; // InfoMessage

procedure TForm1.AddToAnMemo( What : string );
begin
  AnMemo.Text := What + LINE_END + AnMemo.Text;
end; // AddToAnMemo

procedure ChangeBackgroundImage;
begin
  if FileExists( PictureString )
     then begin
       Form1.Image1.Picture.LoadFromFile( PictureString );
       BookEditor.Image1.Picture.LoadFromFile( PictureString );
       SearchEditor.Image1.Picture.LoadFromFile( PictureString );
       EvalForm.Image1.Picture.LoadFromFile( PictureString );
     end;
end; // ChangeBackgroundImage

  {$I prep.pas}      // preparatory procedures
  {$I basic.pas}     // simple and frequently used procedures
  {$I hist.pas}      // a class encapsulating history table
  {$I parse.pas}     // parsing the text strings/files
  {$I disp.pas}      // writing the board in the text mode
  {$I stat.pas}      // class responsible for statistics
  {$I hash.pas}      // hash table routines
  {$I attack.pas}    // testing if a square is attacked
  {$I gen.pas}       // move generation
  {$I pawn_ev.pas}   // pawn evaluation
  {$I patterns.pas}  // attack pattern evaluation
  {$I endgame.pas}   // special endgame evaluation
  {$I eval.pas}      // general evaluation
  {$I recognize.pas} // draw claim and interior node recognizers
  {$I moves.pas}     // make/unmake move
  {$I fen.pas}       // creating FEN strings
  {$I book.pas}      // opening book handling
  {$I quiesc.pas}    // quiescence search
  {$I search.pas}    // main search
  {$I root.pas}      // root search
  {$I iterate.pas}   // iterative deepening
  {$I test.pas}      // test routines
  {$I timing.pas}    // deciding about thinking time

procedure InitBoard; // to be replaced after a FEN reader is created
begin
  ClearBoard;
  Want_resign := 0;

  FillSqr( KING,   E1, 1 );       FillSqr( KING,   E8, -1 );
  FillSqr( QUEEN,  D1, 2 );       FillSqr( QUEEN,  D8, -2 );
  FillSqr( ROOK,   A1, 3 );       FillSqr( ROOK,   A8, -3 );
  FillSqr( ROOK,   H1, 4 );       FillSqr( ROOK,   H8, -4 );
  FillSqr( BISH,   C1, 5 );       FillSqr( BISH,   C8, -6 );
  FillSqr( BISH,   F1, 6 );       FillSqr( BISH,   F8, -5 );
  FillSqr( KNIGHT, B1, 7 );       FillSqr( KNIGHT, B8, -7 );
  FillSqr( KNIGHT, G1, 8 );       FillSqr( KNIGHT, G8, -8 );
  FillSqr( PAWN,   A2, 9 );       FillSqr( PAWN,   A7, -9 );
  FillSqr( PAWN,   B2, 10 );      FillSqr( PAWN,   B7, -10 );
  FillSqr( PAWN,   C2, 11 );      FillSqr( PAWN,   C7, -11 );
  FillSqr( PAWN,   D2, 12 );      FillSqr( PAWN,   D7, -12 );
  FillSqr( PAWN,   E2, 13 );      FillSqr( PAWN,   E7, -13 );
  FillSqr( PAWN,   F2, 14 );      FillSqr( PAWN,   F7, -14 );
  FillSqr( PAWN,   G2, 15 );      FillSqr( PAWN,   G7, -15 );
  FillSqr( PAWN,   H2, 16 );      FillSqr( PAWN,   H7, -16 );

  // set material counters - to be moved into FillSqr
  Board.Index[WPCount] := 8;    Board.Index[BPCount] := 8;   // - to be mo
  Board.Index[WNCount] := 2;    Board.Index[BNCount] := 2;
  Board.Index[WBCount] := 2;    Board.Index[BBCount] := 2;
  Board.Index[WRCount] := 2;    Board.Index[BRCount] := 2;
  Board.Index[WQCount] := 1;    Board.Index[BQCount] := 1;
  Board.Index[WMat]    := -5;   Board.Index[BMat]    := -5;

  GameStack[ 0 ] := Board;  // remember the starting position
end; // InitBoard

procedure SafeUndoMove;
begin
  if StackIndex > 0
     then UnMakeMove;
end; // SafeUndoMove

procedure SafeRedoMove;
begin
   if StackIndex < StackBound
      then begin
        Inc( StackIndex );
        Board := GameStack[ StackIndex ];
      end;
end; // SafeRedoMove

function TForm1.GameToSan : string;
         var A : Integer;
begin
   Result := '';
   for A := 1 to StackIndex do
   begin
     Result := Result + ' ' + MoveToStr( MoveStack[ A ] );
     if A mod 2 = 0 then Result := Result + LINE_END;
   end;
end; // GameToSan

function GetPieceLetter( Sqr : Integer ) : string;
begin
   if Board.List[ Board.Index[Sqr] ].Sqr <> OUTSIDE
      then Result := WPieceChar[ Board.List[Board.Index[Sqr] ].Kind ]
      else Result := '';

   if Result = 'P'
      then Result := '';
end; // GetPieceLetter

function TForm1.MoveToPGN( Move : TMove ) : string;
         var MoveType, Disambiguate : string;
         A : Integer;
begin
   if  ( IsEmpty( Move.SqTo ) )
   and ( Move.Flag <> EPCapt )
      then MoveType := ''
      else begin
        MoveType := 'x';
        if IsPiece( Board.Side, PAWN, Move.SqFr )
           then MoveType := SqrName[ Move.SqFr ][ 1 ] + 'x';
      end;
                           // Disambiguate PGN move in case
   Disambiguate := '';     // when two identical pieces can
   GenMoves;                  // move to single square.
   for A := 1 to MoveNo do
    if  ( MoveSet[A].SqTo =  Move.SqTo )
    and ( MoveSet[A].SqFr <> Move.SqFr )
    and ( Board.List[ Board.Index[Move.SqFr] ].Kind <> PAWN )
    and ( ( Board.List[ Board.Index[Move.SqFr] ].Kind = Board.List[ Board.Index[MoveSet[A].SqFr] ].Kind ) )
    then begin
      if SqrName[Move.SqFr] [1] <> SqrName[MoveSet[A].SqFr] [1]
         then Disambiguate := SqrName[Move.SqFr] [1]
         else Disambiguate := SqrName[Move.SqFr] [2];
      end;

   Result := GetPieceLetter( Move.SqFr )
           + Disambiguate
           + MoveType
           + SqrName[ Move.SqTo ];

   if Move.Flag = CASTLE              // write castling
      then begin                      // move correctly
        if Move.SqTo mod 16 = COL_G
           then Result := '0-0'
           else Result := '0-0-0';
      end;
end; // MoveToPGN

function TForm1.GameToPGN : string;
         var A, LastStackIndex : Integer;
begin
   Result := '';
   LastStackIndex := StackIndex;
   StackIndex := 1;
   SafeUndoMove;
   for A := 1 to LastStackIndex do
   begin // loop
      if ( A mod 2 = 1 )
         then Result := Result + IntToStr( A div 2 + 1 ) + '.';
      Result := Result + MoveToPGN( MoveStack[ A ] ) + ' ';
      MakeMove( MoveStack[A] );
   end; // loop
end; // GameToPGN

procedure SwitchMoveColor;
begin
  ClockSide := ClockSide xor 1;
  if ClockSide = WHITE
     then WMTime := 0
     else BMTime := 0;
end; // SwitchMoveColor

procedure TForm1.BeforeActualMove( Move : TMove );
begin
   Form1.InfoMemo.Font.Size := 8;
   if Notation = N_PGN
      then GameMemo.Text := GameToPGN;
   if Board.Side = WHITE
      then GameMemo.Text := GameMemo.Text + IntToStr( StackIndex div 2 + 1) + '.';
   GameMemo.Text := GameMemo.Text + MoveToPGN( Move );
end; // BeforeActualMove

procedure TForm1.AfterActualMove( Move : TMove );
          var AddStr : string;    // Used whenever a move
begin                             // is made on the board,
   ClaimResult;                   // not during the search.
   RepStack[ StackIndex ] := InitHashValue;
   AddStr := MoveToStr( Move );
   if Board.Side = WHITE
      then AddStr := AddStr + LINE_END;
   {$IFNDEF XBOARD}
   if Notation = N_SAN
      then GameMemo.Text := GameMemo.Text + ' ' + AddStr;
   {$ENDIF}
   SwitchMoveColor;
   StackBound := StackIndex;
end; // AfterActualMove

procedure TForm1.SetProgSide;
begin
  ProgSide := Board.Side;

  if ProgSide = WHITE
     then begin
       WhiteAttack   := ProgAttack;
       WhiteMobility := ProgMobility;
       WhiteShield   := ProgShield;
       BlackAttack   := OppoAttack;
       BlackMobility := OppoMobility;
       BlackShield   := OppoShield;
     end
     else begin
       BlackAttack   := ProgAttack;
       BlackMobility := ProgMobility;
       BlackShield   := ProgShield;
       WhiteAttack   := OppoAttack;
       WhiteMobility := OppoMobility;
       WhiteShield   := OppoShield;
     end;
end; // SetProgSide

function GetOpeningMove : Boolean;
begin
  Result := False;
  if MissedOpening > 5
     then Exit;

  BookNo := 0;
  if not MyBook.IsFlag( MyBook.BookMoveString )
     then begin
       MyMove := OpeningMove;
       Result := True;
     end
     else begin
       BookNo := 1;
       if not MyBook.IsFlag( MyBook.BookMoveString )
          then begin
            MyMove := OpeningMove;
            Result := True;
          end;
     end;
end; // GetOpeningMove

procedure DoThinking;
begin
  ProgThinking := True; // set flag blocking certain user actions
  Form1.SetProgSide;    // initialize asymmetric eval parameters
  UpdateTime;

  if not GetOpeningMove
     then begin
        Inc( MissedOpening );
        MySearch.Iterate( SearchDepth );
     end;

  {$IFNDEF XBOARD}
  Form1.BeforeActualMove( MyMove );
  Form1.GameMemo.Refresh;
  {$ENDIF}

  MakeMove( MyMove );

  {$IFDEF XBOARD}
  Form1.WBThread.SendMove( MoveToStr( MyMove ) );
  {$ENDIF}

  Form1.AfterActualMove( MyMove );
  ProgThinking := False;
end; // DoThinking

function ParseMove( Command : string; FromBoard : Boolean ) : Boolean;
          var MoveRead : TMove;
begin
  Result := False;
  MoveRead := StrToMove( Command );
  if MoveRead.SqFr <> OUTSIDE
     then begin
       Result := True;
       if ( MoveRead.Flag = PROM )
       or ( MoveRead.Flag = CAPTPROM )
       then begin
         MoveRead.Prom := QUEEN;
         if FromBoard = False
            then begin
              if Length( Command ) > 4
                 then begin // perhaps opponent underpromotes
                   case Command[ 5 ] of
                    'n' : MoveRead.Prom := KNIGHT;
                    'r' : MoveRead.Prom := ROOK;
                    'b' : MoveRead.Prom := BISH;
                   end; // of case
                 end;
            end
            else begin // read promotion in graphic mode
              PromForm.Show;
              PromForm.Picked := False;
              repeat Application.ProcessMessages
              until  PromForm.Picked = True;
              MoveRead.Prom := PromForm.PromPiece;
            end;
       end;
       {$IFNDEF XBOARD} Form1.BeforeActualmove( MoveRead ); {$ENDIF}
       MakeMove( MoveRead );
       Form1.AfterActualMove( MoveRead );
     end;
end; // ParseMove

procedure ActivateProgram;
begin
  ProgPassive := False;
  ProgSide    := Board.Side;
  DoThinking;
end; // ActivateProgram

procedure TForm1.Execute( Command : string );
begin
  if  ( AnsiContainsStr( Command, 'go' ) )
  and ( StackIndex <> 2 )
     then ActivateProgram;
  if AnsiContainsStr( Command, 'new' )
     then begin
       PrepareAll;
       History.ClearTable;
       InitBoard;
     end;
  if AnsiContainsStr( Command, 'level ')
     then SetLevel( Command );
  if AnsiContainsStr( Command, 'list' )
     then GenTest;
  if  ( AnsiContainsStr( Command, 'sd ' ) )
  and ( ExtractNumStr( Command ) <> '' )
       then begin
         SearchDepth := StrToInt( ExtractNumStr( Command ) );
         MoveTime := 12 * 3600;
         Timing := FIXED_DEPTH;
       end;  // "sd" command
  if  ( AnsiContainsStr( Command, 'st ' ) )
  and ( ExtractNumStr( Command ) <> '' )
      then begin
         MoveTime := StrToInt( ExtractNumStr(  Command ) );
         SearchDepth := 28;
         Timing := SEC_PER_MOVE;
      end; // "st" command
  if AnsiContainsStr( Command, 'stest' )
     then SpeedTest;
  if AnsiContainsStr( Command, 'stop' )
     then ProgPassive := True;
  if AnsiContainsStr( Command, 'undo' )
     then SafeUndoMove;
  if AnsiContainsStr( Command, 'quit' )
     then Application.Terminate;
  if  ( Length( Command ) >= 4 ) // check if user has entered a move
  and ( ParseMove( Command, False ) )
  and ( not ProgPassive )
       then DoThinking;
end; // Execute

{$IFDEF XBOARD}
procedure TForm1.WinboardMessage( var Msg : TMessage );
begin
  Execute( WbThread.WinboardCommand );
end; // WinBoardMessage - processing message form WBUnit
{$ENDIF}

procedure TForm1.UpdateWindowOptions;
begin
  AnMemo.Color   := WindowBgCol;
  GameMemo.Color := WindowBgCol;
  InfoMemo.Color := WindowBgCol;
  AnMemo.Font.Color   := WindowFontCol;
  GameMemo.Font.Color := WindowFontCol;
  InfoMemo.Font.Color := WindowFontCol;
end; // UpdateWindowOptions

procedure LoadBook;
begin
  BookNo := 0;
  BookName := GuideBookName;
  MyBook.ReadBookFile;
  BookNo := 1;
  BookName := MainBookName;
  MyBook.ReadBookFile;
end; // LoadBook

procedure TForm1.PolishLanguage;
begin
  Game1.Caption     := 'Gra';
  iming1.Caption    := 'Czas';
  Viev1.Caption     := 'Widok';
  Settings1.Caption := 'Ustawienia';
  ools1.Caption     := 'Narzdzia';
  New1.Caption      := 'Nowa';
  Deactivate1.Caption := 'Stop';
  hink1.Caption     := 'Myl';
  Quit1.Caption     := 'Koniec';
  Depth1.Caption    := 'Gboko';
  PerMove1.Caption  := 'Czas na ruch';
  Pergame1.Caption  := 'Czas na gr';
  Invert1.Caption   := 'Obr';
  Colors1.Caption   := 'Kolory';
  Statistics1.Visible := False;
  BoardBackground1.Caption := 'To';
  LightSquared1.Caption := 'Biae pola';
  DarkSquares1.Caption  := 'Czarne pola';
  WhitePieces1.Caption  := 'Biae figury';
  BlackPieces1.Caption  := 'Czarne figury';
  WindowFont1.Caption   := 'Tekst w oknie';
  WindowBackground1.Caption := 'Okna';
  LightFont1.Caption := 'Jasny tekst';
  Darkfont1.Caption  := 'Ciemny tekst';
  BookEditor1.Caption := 'Ksika otwar';
  EvalEditor1.Caption := 'Ocena pozycji';
  SearchEditor1.Caption := 'Parametry przeszukiwania';
  DumpTree1.Caption := 'Analiza drzewa';
  Playername1.Caption := 'Nazwisko zawodnika';
  SaveCurrent1.Caption := 'Zapamitaj ustawienia';
  RestoreOriginal1.Caption := 'Ustawienia fabryczne';

  BookStyle1.Caption := 'Styl otwarcia';
  Hypermodern1.Caption := 'Hipermodernistyczny';
  Broad1.Caption := 'Gra wszystko';
  Solid1.Caption := 'Solidny';
  Gambit1.Caption := 'Gambitowy';
  Normal1.Caption := 'Normalny';
  Fighter1.Caption := 'Aktywny';
  Unusual1.Caption := 'Dziwny';
  BookOff1.Caption := 'Bez ksiki';

  PlayingStyle1.Caption := 'Styl gry';
  Smart1.Caption := 'Sprytny';
  Objective1.Caption := 'Obiektywny';
  Pessimist1.Caption := 'Pesymista';

  Lauouts1.Caption := 'Grafika';
  YellowWater1.Caption := 'Zota woda';
  Glacier1.Caption := 'Lodowiec';
  Wood1.Caption := 'Ciemne drewno';
  Subdued1.Caption := 'Wyciszony';
  Language1.Caption := 'Jzyk';
  WPName.Caption := 'Biae';
  BPName.Caption := 'Czarne';

  NewGameButt.Hint := 'Nowa gra';       // hints for buttons
  InvertButt.Hint := 'Obr szachownic';
  StopButt.Hint := 'Program nie gra';
  BackAllButt.Hint := 'Cofnij wszystko';
  BackButt.Hint := 'Cofnij ruch';
  FwdButt.Hint := 'Powtrz ruch';
  FwdAllButt.Hint := 'Id do koca';
  ThinkButt.Hint  := 'Graj';
  EvalButt.Hint := 'Ocena pozycji';
  BookButt.Hint := 'Ksika otwar';
  AnalyzeButt.Hint := 'Poka/schowaj analiz';
  StatButt.Hint := 'Statystyki';
  InfoButt.Hint := 'O programie';
  QuitButt.Hint := 'Koniec';

  sProgramInfo := 'Clericus 1.03' + LINE_END
                       + 'napisa Pawel Koziol'
                       + LINE_END + 'nescitus@o2.pl';
  sDrawByRep   := 'Remis' + LINE_END + 'pezez powtrzenie';
  sDrawBy50m   := 'Remis' + LINE_END + 'zasada 50 posuni';
  sDrawByMat   := 'Remis' + LINE_END + 'niewystarczajcy materia';
  sCheckmate   := 'Mat';
  sStalemate   := 'Pat';
  sBookSaving  := 'Zapisywanie ksiki...';
  sBookSaved   := 'Ksika zapisana';
  sMoveBack    := 'Cofanie';
  sMoveFwd     := 'Przywracanie';
  sFwdAll      := 'Koniec partii';
  sBackAll     := 'Pocztek partii';
end; // Polish language

procedure TForm1.EnglishLanguage;
begin
  Game1.Caption  := 'Game';
  iming1.Caption := 'Timing';
  Viev1.Caption  := 'Viev';
  Settings1.Caption := 'Settings';
  ools1.Caption  := 'Tools';
  New1.Caption   := 'New';
  Deactivate1.Caption := 'Deactivate';
  hink1.Caption  := 'Think';
  Quit1.Caption  := 'Quit';
  Depth1.Caption := 'Depth';
  PerMove1.Caption := 'Time per move';
  Pergame1.Caption := 'Time per game';
  Invert1.Caption := 'Invert';
  Colors1.Caption := 'Colors';
  Statistics1.Visible := True;
  BoardBackground1.Caption := 'Board Background';
  LightSquared1.Caption := 'Light squares';
  DarkSquares1.Caption  := 'Dark squares';
  WhitePieces1.Caption  := 'White pieces';
  BlackPieces1.Caption  := 'Black pieces';
  WindowFont1.Caption   := 'Window font';
  WindowBackground1.Caption := 'Window background';
  LightFont1.Caption := 'Light font';
  Darkfont1.Caption  := 'Dark font';
  BookEditor1.Caption := 'Book editor';
  EvalEditor1.Caption := 'Eval editor';
  SearchEditor1.Caption := 'Search editor';
  DumpTree1.Caption := 'Dump tree';
  Playername1.Caption := 'Player name';
  SaveCurrent1.Caption := 'Save current';
  RestoreOriginal1.Caption := 'Restore original';

  BookStyle1.Caption   := 'Book style';
  Hypermodern1.Caption := 'Hypermodern';
  Broad1.Caption       := 'Broad';
  Solid1.Caption       := 'Solid';
  Gambit1.Caption      := 'Gambit';
  Normal1.Caption      := 'Normal';
  Fighter1.Caption     := 'Fighter';
  Unusual1.Caption     := 'Unusual';
  BookOff1.Caption     := 'Book off';

  PlayingStyle1.Caption := 'Playing style';
  Smart1.Caption        := 'Smart';
  Objective1.Caption    := 'Objective';
  Pessimist1.Caption    := 'Pessimist';

  Lauouts1.Caption     := 'Layouts';
  YellowWater1.Caption := 'Yellow water';
  Glacier1.Caption     := 'Glacier';
  Wood1.Caption        := 'Wood';
  Subdued1.Caption     := 'Subdued';
  Language1.Caption    := 'Language';
  WPName.Caption       := 'White';
  BPName.Caption       := 'Black';

  NewGameButt.Hint := 'New game';      // hints for buttons
  InvertButt.Hint  := 'Turn board';
  StopButt.Hint    := 'Deactivate engine';
  BackAllButt.Hint := 'Undo all moves';
  BackButt.Hint    := 'Undo move';
  FwdButt.Hint     := 'Redo move';
  FwdAllButt.Hint  := 'Redo all moves';
  ThinkButt.Hint   := 'Compute';
  EvalButt.Hint    := 'Evaluation details';
  BookButt.Hint    := 'Book editor';
  AnalyzeButt.Hint := 'Show/hide analysis';
  StatButt.Hint    := ' Search statistics';
  InfoButt.Hint    := 'About';
  QuitButt.Hint    := 'Quit';

  sProgramInfo := 'Clericus 1.03' + LINE_END
                       + 'a chess program by Pawel Koziol'
                       + LINE_END + 'nescitus@o2.pl';
  sDrawByRep   := 'Draw' + LINE_END + 'by repetition';
  sDrawBy50m   := 'Draw' + LINE_END + '50 move rule';
  sDrawByMat   := 'Draw' + LINE_END + 'insufficient material';
  sCheckmate   := 'Checkmate!';
  sStalemate   := 'Stalemate!';
  sBookSaving  := 'Saving book...';
  sBookSaved   := 'Book saved';
  sMoveBack    := 'Taking back';
  sMoveFwd     := 'Replaying';
  sFwdAll      := 'Go to end';
  sBackAll     := 'Go to start';
end; // EnglishLanguage

procedure TForm1.LanguageSettings;
begin
  case Language of
       POLISH  : PolishLanguage;
       ENGLISH : EnglishLanguage;
  end; // of case
end; // LanguageSettings

procedure TForm1.BoardGridDrawCell( Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState );
          var S : string;
begin
   with BoardGrid do begin
     S := Cells[ ACol, ARow ];

     // Fill rectangle with colour
     if BoardSqrCol[ ACol, ARow ] = WHITE
        then Canvas.Brush.Color := LtSqrCol
        else Canvas.Brush.Color := DarkSqrCol;
     Canvas.FillRect( Rect );

     // Next, draw the text in the rectangle
     if PieceCol[ ACol, ARow ] = WHITE
        then Canvas.Font.Color := WhitePieceCol
        else Canvas.Font.Color := BlackPieceCol;
     Canvas.TextOut( Rect.Left + 2, Rect.Top + 2, S );
   end; // with BoardGrid
end;    // BoardGridDrawCell

procedure TForm1.InvertPieceStruct;
          var A, B     : Integer;
          TempIntArray : TDispArray;
          TempStrArray : array[ 0..7, 0..7 ] of string;
begin
  for A := 0 to 7 do
  for B := 0 to 7 do
  begin
    TempIntArray[ A, B ] := PieceCol[ 7 - A, 7 - B ];
    TempStrArray[ A, B ] := BoardGrid.Cells[7 - A, 7 - B ];
  end;

  PieceCol := TempIntArray;  // invert BoardGrid display
  for A := 0 to 7 do
  for B := 0 to 7 do
      BoardGrid.Cells[ A, B ] := TempStrArray[ A, B ];
end; // InvertPieceStruct

procedure TForm1.RefreshBoardDisplay;
begin
  FenToGraphBoard( BoardToFen );
  if Invert1.Checked
     then InvertPieceStruct;
  if BookEditor1.Checked
     then BookEditor.RefreshForm;
  BoardGrid.Color := BoardBgColor;
  BoardGrid.Repaint;
  WTime.Font.Color := DarkFontCol;
  BTime.Font.Color := DarkFontCol;
  WMoveTime.Font.Color := LightFontCol;
  BMoveTime.Font.Color := LightFontCol;
  UpdateWindowOptions;
end; // RefreshBoardDisplay

procedure TForm1.BoardGridMouseDown( Sender: TObject; Button: TMouseButton;
   Shift: TShiftState; X, Y: Integer );
          var FenString, MoveString : string;
begin
  if InputMode = imDragging
  then begin
    BoardGrid.MouseToCell( X, Y, StartCol, StartRow );
    if Invert1.Checked
       then begin
         StartCol := 7 - StartCol;
         StartRow := 7 - StartRow;
       end;
  end
  else begin // InputMode = imClicking
    BoardGrid.MouseToCell( X, Y, EndCol, EndRow );
    if Invert1.Checked
       then begin
         EndCol := 7 - EndCol;
         EndRow := 7 - EndRow;
       end;
    FenString  := BoardToFen;

    MoveString := ColName[ StartCol ] + RowName[ StartRow ]
                + ColName[ EndCol ]   + RowName[ EndRow ];
    if ParseMove( MoveString, True )
       then begin
         MoveMade := True;
         if BookEditor.FeedFromBoard1.Checked
            then MyBook.WriteBookmove( FenString, MoveString );
         RefreshBoardDisplay;
         if not ProgPassive
            then begin
              ActivateProgram;
              RefreshBoardDisplay;
            end; // not ProgPassive
    end;
  end;
end; // BoardGridMouseDown

procedure TForm1.BoardGridMouseUp( Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
         var FenString, MoveString : String;
begin
  BoardGrid.MouseToCell( X, Y, EndCol, EndRow );

  if Invert1.Checked
     then begin
       EndCol := 7 - EndCol;
       EndRow := 7 - EndRow;
     end;

  if  ( StartCol = EndCol )
  and ( StartRow = EndRow )
      then begin
        InputMode := imClicking;
        Exit;
      end
      else InputMode := imDragging;

  FenString  := BoardToFen;
  MoveString := ColName[ StartCol ] + RowName[ StartRow ]
              + ColName[ EndCol ]   + RowName[ EndRow ];

  if ParseMove( MoveString, True )
     then begin
       MoveMade := True;
       if BookEditor.FeedFromBoard1.Checked
          then MyBook.WriteBookmove( FenString, MoveString );
       RefreshBoardDisplay;
       if not ProgPassive
          then begin
            ActivateProgram;
            RefreshBoardDisplay;
          end; // not ProgPassive
     end;
end; // BoardGridMouseUp

procedure PliesUncheck;
begin
  Form1.N4Plies1.Checked := True;
  Form1.N4Plies1.Checked := False;
end; // PliesUncheck

procedure SecondsUncheck;
begin
  Form1.N60sec1.Checked := True;
  Form1.N60sec1.Checked := False;
end; // SecondsUncheck

procedure TimeUncheck;
begin
  Form1.N2h1.Checked := True;
  Form1.N2h1.Checked := False;
end; // GameUncheck

procedure TForm1.SetSeconds( Ex : string );
begin
  PliesUncheck;
  TimeUncheck;
  Execute( Ex );
end; // SetSeconds

procedure TForm1.SetDepth( Ex : string );
begin
  TimeUncheck;
  SecondsUncheck;
  Execute( Ex );
end; // SetDepth

procedure TForm1.SetTime( Ex : string );
begin
  SecondsUncheck;
  PliesUncheck;
  Execute( Ex );
end; // SetTime

procedure TForm1.N120s1Click(Sender: TObject);
begin
  N120s1.Checked := True;
  SetSeconds( 'st 120' );
end;

procedure TForm1.N60sec1Click( Sender: TObject );
begin
  N60sec1.Checked := True;
  SetSeconds( 'st 60' );
end;

procedure TForm1.N30sec1Click( Sender: TObject );
begin
  N30sec1.Checked := True;
  SetSeconds( 'st 30' );
end;

procedure TForm1.N10sec1Click( Sender: TObject );
begin
  N10sec1.Checked := True;
  SetSeconds( 'st 10' );
end;

procedure TForm1.N5sec1Click( Sender: TObject );
begin
  N5sec1.Checked := True;
  SetSeconds( 'st 5' );
end;

procedure TForm1.N1sec1Click( Sender: TObject );
begin
  N1Sec1.Checked := True;
  SetSeconds( 'st 1' );
end;

procedure TForm1.N4plies1Click( Sender: TObject );
begin
  N4Plies1.Checked := True;
  SetDepth( 'sd 4');
end;

procedure TForm1.N5plies1Click( Sender: TObject );
begin
  N5Plies1.Checked := True;
  SetDepth( 'sd 5' );
end;

procedure TForm1.N6plies1Click( Sender: TObject );
begin
  N6Plies1.Checked := True;
  SetDepth( 'sd 6' );
end;

procedure TForm1.N7plies1Click( Sender: TObject );
begin
  N7Plies1.Checked := True;
  SetDepth( 'sd 7' );
end;

procedure TForm1.N8plies1Click( Sender: TObject );
begin
  N8Plies1.Checked := True;
  SetDepth( 'sd 8' );
end;

procedure TForm1.N9plies1Click( Sender: TObject );
begin
  N9Plies1.Checked := True;
  SetDepth( 'sd 9' );
end;

procedure TForm1.N10plies1Click( Sender: TObject );
begin
  N10plies1.Checked := True;
  SetDepth( 'sd 10' );
end;

procedure TForm1.N12plies1Click(Sender: TObject);
begin
  N12plies1.Checked := True;
  SetDepth( 'sd 12' );
end;

procedure TForm1.N5min1Click( Sender: TObject );
begin
  N5min1.Checked := True;
  SetTime( 'level 0 5 0' );
end;

procedure TForm1.N10min1Click( Sender: TObject );
begin
  N10Min1.Checked := True;
  SetTime( 'level 0 10 0' );
end;

procedure TForm1.N15min1Click( Sender: TObject );
begin
  N15min1.Checked := True;
  SetTime( 'level 0 15 0' );
end;

procedure TForm1.N30min1Click( Sender: TObject );
begin
  N30min1.Checked := True;
  SetTime( 'level 0 30 0' );
end;

procedure TForm1.N1h1Click( Sender: TObject );
begin
  N1h1.Checked := True;
  SetTime( 'level 0 60 0' );
end;

procedure TForm1.N15h1Click( Sender: TObject );
begin
  N15h1.Checked := True;
  SetTime( 'level 0 90 0' );
end;

procedure TForm1.N2h1Click( Sender: TObject );
begin
  N2h1.Checked := True;
  SetTime( 'level 0 120 0' );
end;

procedure TForm1.New1Click( Sender: TObject );
begin
  if ProgThinking
     then Exit;
  Execute( 'new' );
  FenToGraphBoard( BoardToFen );
  if BookEditor1.Checked
     then begin
       ProgPassive := True;
       BookEditor.RefreshForm;
     end;
end; // New1Click

procedure TForm1.Quit1Click( Sender: TObject );
begin
  SetSeconds( 'st 1' );
  Application.Terminate;
end; // Quit1Click

procedure TForm1.Statistics1Click( Sender: TObject );
begin
  AnMemo.Text := MyStat.MakeString;
end; // Statistics1Click

procedure TForm1.hink1Click( Sender: TObject );
begin
  InfoMemo.Font.Size := 8;
  if ProgThinking
     then Exit;
  ActivateProgram;
  RefreshBoardDisplay;
end; // Think1Click

procedure TForm1.Invert1Click( Sender: TObject );
begin
  if Invert1.Checked then Invert1.Checked := False
                     else Invert1.Checked := True;
  InvertPieceStruct;
end; // Invert1Click

procedure TForm1.Deactivate1Click( Sender: TObject );
begin
   Execute( 'stop' );
end; // Deactivate1Click

procedure TForm1.Lightsquared1Click( Sender: TObject );
begin
	if ColorDialog.Execute
  then begin
    LtSqrCol := ColorDialog.Color;
    RefreshBoardDisplay;
  end;
end; // LightSquares1Click

procedure TForm1.Darksquares1Click( Sender: TObject );
begin
	if ColorDialog.Execute
  then begin
    DarkSqrCol := ColorDialog.Color;
    RefreshBoardDisplay;
  end;
end; // DarkSquares1Click

procedure TForm1.Whitepieces1Click( Sender: TObject );
begin
	if ColorDialog.Execute
  then begin
    WhitePieceCol := ColorDialog.Color;
    RefreshBoardDisplay;
  end;
end; // WhitePieces1Click

procedure TForm1.Blackpieces1Click( Sender: TObject );
begin
	if ColorDialog.Execute
  then begin
    BlackPieceCol := ColorDialog.Color;
    RefreshBoardDisplay;
  end;
end; // BlackPieces1Click

procedure TForm1.Timer1Timer( Sender: TObject );
begin
  if ClockSide = WHITE
     then begin
       WPName.Font.Color := DarkFontCol;
       BPName.Font.Color := LightFontCol;
       WhiteClockBevel.Style := BsRaised;
       BlackClockBevel.Style := BSLowered;
       Inc( WMTime );
       WMoveTime.Caption := SecToClock( WMTime );
       Inc( WTotTime );
       WTime.Caption := SecToClock( WTotTime );
     end
     else begin
       BPName.Font.Color := DarkFontCol;
       WPName.Font.Color := LightFontCol;
       WhiteClockBevel.Style := BsLowered;
       BlackClockBevel.Style := BSRaised;
       Inc( BMTime );
       BMoveTime.Caption := SecToClock( BMTime );
       Inc( BTotTime );
       BTime.Caption := SecToClock( BTotTime );
     end;
end; // Form1.Timer1Timer

procedure TForm1.UpdateNotation;
begin
  if Notation = N_SAN
     then GameMemo.Text := GameToSan
     else GameMemo.Text := GameToPGN;
end; // UpdateNotation

procedure TForm1.BackButtClick( Sender: TObject );
begin
  if ProgThinking
     then Exit;
  InfoMemo.Font.Size := 8;
  InfoMemo.Text := sMoveBack;
  SafeUndoMove;
  UpdateNotation;
  RefreshBoardDisplay;
  SwitchMoveColor;
end; // undoing a move

procedure TForm1.BackAllButtClick( Sender: TObject );
begin
  if ProgThinking
     then Exit;
  InfoMemo.Font.Size := 8;
  InfoMemo.Text := 'Back to start';
  StackIndex := 1;
  SafeUndoMove;
  InfoMemo.Text := sBackAll;
  UpdateNotation;
  RefreshBoardDisplay;
  ClockSide := Board.Side;
end; // undoing all moves

procedure TForm1.FwdButtClick( Sender: TObject );
begin
  if ProgThinking
     then Exit;
  InfoMemo.Font.Size := 8;
  InfoMemo.Text := sMoveFwd;
  SafeRedoMove;
  UpdateNotation;
  RefreshBoardDisplay;
  SwitchMoveColor;
end; // redoing a move

procedure TForm1.FwdAllButtClick( Sender: TObject );
begin
  if ProgThinking
     then Exit;
  InfoMemo.Font.Size := 8;
  InfoMemo.Text := '';
  StackIndex := StackBound - 1;
  SafeRedoMove;
  UpdateNotation;
  RefreshBoardDisplay;
  ClockSide := Board.Side;
  InfoMemo.Text := sFwdAll;
end; // going to the end of the game

procedure TForm1.BookEditor1Click( Sender: TObject );
begin
  if BookEditor1.Checked
     then begin
       BookEditor1.Checked := False;
       BookEditor.FeedFromBoard1.Checked := False;
       BookEditor.Hide;
     end
     else begin
       BookEditor1.Checked := True;
       ProgPassive := True;
       BookEditor.Show;
     end;
end; // opens or closes book editor

procedure TForm1.AnalyzeButtClick( Sender: TObject );
begin
  if AnMemo.Visible = True
     then begin
       AnMemo.Hide;
       InfoMemo.Hide;
       Height := 493;
     end
     else begin
       AnMemo.Show;
       InfoMemo.Show;
       Height := 593;
     end;
end; // shows or hides analysis window

procedure TForm1.InfoButtClick( Sender: TObject );
begin
  ShowMessage( sProgramInfo );
end; // displays program info

procedure TForm1.EvalButtClick( Sender: TObject );
begin
  if ProgThinking
     then Exit;
  if DebugForm.Visible = False
     then begin
       SetProgSide;
       DebugForm.GeneralMemo.Text := MyEval.EvalString;
       DebugForm.WhiteMemo.Text   := MyEval.DebugWhite;
       DebugForm.BlackMemo.Text   := MyEval.DebugBlack;
       DebugForm.Show;
     end
     else DebugForm.Hide;  
end; // displays evaluation details

procedure TForm1.SaveCurrent1Click( Sender: TObject );
begin
  WriteIniFile( 'clericus.ini' );
end; // saves current settings

procedure TForm1.RestoreOriginal1Click( Sender: TObject );
begin
  RestoreSettings;
  BoardGrid.Repaint;
end; // restores original settings

procedure TForm1.LightFont1Click( Sender: TObject );
begin
	if ColorDialog.Execute
  then begin
    LightFontCol := ColorDialog.Color;
    RefreshBoardDisplay;
  end;
end; // LightFont1Click

procedure TForm1.DarkFont1Click( Sender: TObject );
begin
	if ColorDialog.Execute
  then begin
    DarkFontCol := ColorDialog.Color;
    RefreshBoardDisplay;
  end;
end; // DarkFont1Click

procedure TForm1.WindowFont1Click( Sender: TObject );
begin
	if ColorDialog.Execute
  then begin
    WindowFontCol := ColorDialog.Color;
    RefreshBoardDisplay;
  end;
end; // WindowFont1Click

procedure TForm1.WindowBackground1Click( Sender: TObject );
begin
	if ColorDialog.Execute
  then begin
    WindowBgCol := ColorDialog.Color;
    RefreshBoardDisplay;
  end;
end; // WindowBackground1Click

procedure TForm1.PlayerName1Click( Sender: TObject );
begin
   NameDialog.IsInput := False;
   NameDialog.Show;
   repeat Application.ProcessMessages
   until  NameDialog.Visible = False;
   if NameDialog.IsInput
      then begin
        PlayerName := NameDialog.NameString;
      end;
end; // PlayerName1Click

procedure TForm1.EvalEditor1Click( Sender: TObject );
begin
  EvalForm.Show;
end; // EvalEditor1Click

procedure TForm1.BoardBackground1Click(Sender: TObject);
begin
	if ColorDialog.Execute
  then begin
    BoardBgColor := ColorDialog.Color;
    RefreshBoardDisplay;
  end;
end; // BoardBackground1Click

procedure TForm1.Objective1Click( Sender: TObject );
begin
  Objective1.Checked := True;
  ReadIniFile( 'style/objective.ini' );
end; // Objective1Click

procedure TForm1.Smart1Click(Sender: TObject);
begin
  Smart1.Checked := True;
  ReadIniFile( 'style/smart.ini' );
end; // Smart1Click

procedure TForm1.Pessimist1Click(Sender: TObject);
begin
  Pessimist1.Checked := True;
  ReadIniFile( 'style/pessimist.ini' );
end; // Pessimist1Click

procedure TForm1.BookSelection( CurrentBook : string );
begin
  MyBook.ClearBookArray( 0 );
  GuideBookName   :=  CurrentBook;
  LoadBook;
  if BookEditor1.Checked
     then BookEditor.RefreshForm;
end; // BookSelection

procedure TForm1.BookOff1Click(Sender: TObject);
begin
  MissedOpening := 100;
  BookOff1.Checked := True;
end; // BookOff1Click

procedure TForm1.Normal1Click( Sender: TObject );
begin
  Normal1.Checked := True;
  BookSelection( 'books/MainBook.hop' );
  MissedOpening := 0;
end; // Normal1Click

procedure TForm1.Broad1Click( Sender : TObject );
begin
  Broad1.Checked := True;
  BookSelection( 'books/Broad.hop' );
  MissedOpening := 0;
end; // Broad1Click

procedure TForm1.Solid1Click( Sender : TObject );
begin
  Solid1.Checked  :=  True;
  BookSelection( 'books/Solid.hop' );
  MissedOpening := 0;
end; // Solid1Click

procedure TForm1.Hypermodern1Click( Sender : TObject );
begin
  Hypermodern1.Checked := True;
  BookSelection( 'books/Hypermodern.hop' );
  MissedOpening := 0;
end; // Hypermodern1Click

procedure TForm1.Fighter1Click( Sender : TObject );
begin
  Fighter1.Checked := True;
  BookSelection( 'books/Fighter.hop' );
  MissedOpening := 0;
end; // Fighter1Click

procedure TForm1.Gambit1Click( Sender : TObject );
begin
  Gambit1.Checked :=  True;
  BookSelection( 'books/Gambit.hop' );
  MissedOpening := 0;
end; // Gambit1Click

procedure TForm1.Unusual1Click( Sender : TObject );
begin
  Unusual1.Checked :=  True;
  BookSelection( 'books/Unusual.hop' );
  MissedOpening := 0;
end; // Unusual1Click

procedure TForm1.YellowWater1Click( Sender : TObject );
begin
  ReadIniFile( 'graf/yellow.txt' );
  ChangeBackgroundImage;
end; // YellowWater1Click

procedure TForm1.Glacier1Click( Sender : TObject );
begin
  ReadIniFile( 'graf/glacier.txt' );
  ChangeBackgroundImage;
end; // Glacier1Click

procedure TForm1.Wood1Click( Sender : TObject );
begin
  ReadIniFile( 'graf/wood.txt' );
  ChangeBackgroundImage;
end; // Wood1Click

procedure TForm1.undra1Click( Sender : TObject );
begin
  ReadIniFile( 'graf/tundra.txt' );
  ChangeBackgroundImage;
end; // Tundra1Click

procedure TForm1.Subdued1Click( Sender : TObject );
begin
  ReadIniFile('graf/subdued.txt');
  ChangeBackgroundImage;
end; // Subdued1Click

procedure TForm1.Polish1Click( Sender: TObject );
begin
  PolishLanguage;
  EvalForm.PolishLanguage;
  BookEditor.PolishLanguage;
  SearchEditor.PolishLanguage;
end; // Polish1Click

procedure TForm1.English1Click( Sender : TObject );
begin
  EnglishLanguage;
  EvalForm.EnglishLanguage;
  BookEditor.EnglishLanguage;
  SearchEditor.EnglishLanguage;
end; // English1.Click

procedure TForm1.SearchEditor1Click(Sender: TObject);
begin
  SearchEditor.Show;
end; // SearchEditor1Click

procedure TForm1.DumpTree1Click(Sender: TObject);
begin
  IsDumping := True;
  DumpForm.Show;
end; // DumpTree1Click

procedure TForm1.FormResize( Sender: TObject );
begin
  Image1.Width := Width - 10;
  Image1.Height := Height - 50;
end; // FormResize

procedure TForm1.FormClose( Sender: TObject; var Action: TCloseAction );
begin
   SetSeconds( 'st 1' );  // forces immediate exit
end; // FormClose

procedure TForm1.FormCreate;
          var A : Integer;
begin
  for A := 1 to 10 do Randomize;

  // Initialize assymetric values in case ini file does
  // nothing with them.

  ProgAttack    := 200;      OppoAttack    := 200;
  ProgMobility  := 100;      OppoMobility  := 150;
  ProgShield    := 100;      OppoShield    := 100;

  {$IFDEF XBOARD} WbThread := TWbThread.Create( false, Handle );
  WbThread.SendCommand( 'feature done="0"', True ); {$ENDIF}
    SetHashData;
    MyStat   := TStat.Create;
    History  := THistory.Create;
    MyEval   := TEval.Create;
    MyBook   := TBook.Create;
    MySearch := TSearch.Create;
    ReadIniFile( 'clericus.ini' );
    LoadBook;
    MyBook.ReadCommentFile;
    PrepareAll;
    InitBoard;
    ProgPassive := False;
  {$IFDEF XBOARD} WbThread.SendCommand( 'feature done="1"', True );
  {$ELSE}
  FenToGraphBoard( BoardToFen );
  AnMemo.Text := '';
  GameMemo.Text := '';
  InfoMemo.Text := '';
  if FileExists( PictureString )
     then Image1.Picture.LoadFromFile( PictureString );
  UpdateWindowOptions;
  LanguageSettings;
  Top  := 1;
  Left := 1;
  Repaint;
  {$ENDIF}
end;  // TForm1.FormCreate

procedure TForm1.FormDestroy( Sender: TObject );
begin
  MyStat.Destroy;
  History.Destroy;
  MyBook.Destroy;
  MyEval.Destroy;
  MySearch.Destroy;
end; // FormDestroy

procedure TForm1.Genspeed1Click(Sender: TObject);
          var A : Integer;
begin
   STi := Time;
   for A := 1 to 10000000 do GenMoves;
   ETi := Time;
   ShowMessage( '10.000.000 move generations in ' + SecToClock( SecondsBetween( ETi, STi ) ) );
end; // GenSpeed

end.
